from typing import Tuple

from PIL import Image

from nyckel.config import MAX_IMAGE_SIZE_PIXELS


class ImageResizer:

    def __init__(self, max_image_size_pixels: int = MAX_IMAGE_SIZE_PIXELS):
        self._max_image_size_pixels = max_image_size_pixels

    def __call__(self, img: Image.Image) -> Image.Image:
        if not self._needs_resize(img.width, img.height):
            return img
        new_width, new_height = self._get_new_width_height(img.width, img.height)
        img = img.resize((new_width, new_height))
        return img

    def _needs_resize(self, width: int, height: int) -> bool:
        return width > self._max_image_size_pixels or height > self._max_image_size_pixels

    def _get_new_width_height(self, width: int, height: int) -> Tuple[int, int]:
        if width > height:
            new_width = self._max_image_size_pixels
            new_height = int(new_width * height / width)
        else:
            new_height = self._max_image_size_pixels
            new_width = int(new_height * width / height)
        return new_width, new_height



import os
import time
from typing import Iterator

import numpy as np
import pytest
import requests
from nyckel import (
    ClassificationAnnotation,
    ClassificationFunction,
    ClassificationLabel,
    Credentials,
    ImageClassificationFunction,
    ImageClassificationSample,
    ImageTagsFunction,
    TabularClassificationFunction,
    TabularClassificationSample,
    TextClassificationFunction,
    TextClassificationSample,
    TextTagsFunction,
)
from nyckel.functions.classification.image_classification import ImageEncoder
from PIL import Image


def make_random_image(size: int = 100) -> str:
    imarray = np.random.rand(size, size, 3) * 255
    img = Image.fromarray(imarray.astype("uint8")).convert("RGB")
    return ImageEncoder().to_base64(img)


def hold_until_list_samples_available(function: ClassificationFunction, expected_count: int) -> None:
    actual_count = 0
    while not actual_count == expected_count:
        actual_count = len(function.list_samples())
        time.sleep(0.25)


@pytest.fixture
def auth_test_credentials() -> Iterator[Credentials]:
    credentials = get_test_credentials()
    assert_credentials_has_access(credentials)
    if not credentials_has_project(credentials):
        create_project_for_credentials(credentials)
    yield credentials


@pytest.fixture
def text_classification_function(auth_test_credentials: Credentials) -> Iterator[TextClassificationFunction]:
    func = TextClassificationFunction.create("PYTHON-SDK TEXT TEST FUNCTION", auth_test_credentials)
    yield func
    func.delete()


@pytest.fixture
def text_classification_function_with_content(
    auth_test_credentials: Credentials,
) -> Iterator[TextClassificationFunction]:
    func = TextClassificationFunction.create("PYTHON-SDK TEXT TEST FUNCTION", auth_test_credentials)
    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)
    nice = ClassificationAnnotation(label_name="Nice")
    boo = ClassificationAnnotation(label_name="Boo")
    samples = [
        TextClassificationSample(data="hello", external_id="1", annotation=nice),
        TextClassificationSample(data="hi", external_id="2", annotation=nice),
        TextClassificationSample(data="Good bye", external_id="3", annotation=boo),
        TextClassificationSample(data="I'm leaving", external_id="4", annotation=boo),
        TextClassificationSample(data="Hi again", external_id="5"),
    ]
    func.create_samples(samples)
    hold_until_list_samples_available(func, len(samples))
    yield func
    func.delete()


@pytest.fixture
def image_classification_function(auth_test_credentials: Credentials) -> Iterator[ImageClassificationFunction]:
    func = ImageClassificationFunction.create("PYTHON-SDK IMAGE TEST FUNCTION", auth_test_credentials)
    yield func
    func.delete()


@pytest.fixture
def image_classification_function_with_content(
    auth_test_credentials: Credentials,
) -> Iterator[ImageClassificationFunction]:
    func = ImageClassificationFunction.create("PYTHON-SDK IMAGE TEST FUNCTION", auth_test_credentials)
    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)
    nice = ClassificationAnnotation(label_name="Nice")
    boo = ClassificationAnnotation(label_name="Boo")
    samples = [
        ImageClassificationSample(data=make_random_image(), external_id="1", annotation=nice),
        ImageClassificationSample(data=make_random_image(), external_id="2", annotation=nice),
        ImageClassificationSample(data=make_random_image(), external_id="3", annotation=boo),
        ImageClassificationSample(data=make_random_image(), external_id="4", annotation=boo),
        ImageClassificationSample(data=make_random_image(), external_id="5"),
    ]
    func.create_samples(samples)
    hold_until_list_samples_available(func, len(samples))
    yield func
    func.delete()


@pytest.fixture
def tabular_classification_function(auth_test_credentials: Credentials) -> Iterator[TabularClassificationFunction]:
    func = TabularClassificationFunction.create("PYTHON-SDK TABULAR TEST FUNCTION", auth_test_credentials)
    yield func
    func.delete()


@pytest.fixture
def tabular_classification_function_with_content(
    auth_test_credentials: Credentials,
) -> Iterator[TabularClassificationFunction]:
    func = TabularClassificationFunction.create("PYTHON-SDK TABULAR TEST FUNCTION", auth_test_credentials)
    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)
    nice = ClassificationAnnotation(label_name="Nice")
    boo = ClassificationAnnotation(label_name="Boo")
    samples = [
        TabularClassificationSample(data={"firstname": "Adam", "lastname": "Adams"}, external_id="1", annotation=nice),
        TabularClassificationSample(data={"firstname": "Bo", "lastname": "Biden"}, external_id="2", annotation=nice),
        TabularClassificationSample(data={"firstname": "Carl", "lastname": "Carrot"}, external_id="3", annotation=boo),
        TabularClassificationSample(data={"firstname": "Dick", "lastname": "Denali"}, external_id="4", annotation=boo),
        TabularClassificationSample(data={"firstname": "Erik", "lastname": "Elk"}, external_id="5"),
    ]
    func.create_samples(samples)
    hold_until_list_samples_available(func, len(samples))
    yield func
    func.delete()


@pytest.fixture
def text_tags_function(auth_test_credentials: Credentials) -> Iterator[TextTagsFunction]:
    func = TextTagsFunction.create("PYTHON-SDK TEXT TAGS TEST FUNCTION", auth_test_credentials)
    yield func
    func.delete()


@pytest.fixture
def image_tags_function(auth_test_credentials: Credentials) -> Iterator[ImageTagsFunction]:
    func = ImageTagsFunction.create("PYTHON-SDK IMAGE TAGS TEST FUNCTION", auth_test_credentials)
    yield func
    func.delete()


def get_test_credentials() -> Credentials:
    if "NYCKEL_PYTHON_SDK_CLIENT_SECRET" in os.environ:
        credentials = Credentials(
            client_id="python-sdk-test",
            client_secret=os.environ["NYCKEL_PYTHON_SDK_CLIENT_SECRET"],
            server_url="https://www.nyckel-staging.com",
        )
    else:
        credentials = Credentials(
            client_id="python-sdk-test",
            client_secret="secret",
            server_url="http://localhost:5000",
        )
    return credentials


def assert_credentials_has_access(credentials: Credentials) -> None:
    credentials.token


def credentials_has_project(credentials: Credentials) -> bool:
    session = requests.Session()
    session.headers.update({"authorization": "Bearer " + credentials.token})
    response = session.get(f"{credentials.server_url}/v0.9/projects")
    assert response.status_code == 200
    return len(response.json()) > 0


def create_project_for_credentials(credentials: Credentials) -> None:
    session = credentials.get_session()
    print("-> Creating project for credentials")
    response = session.post(f"{credentials.server_url}/v0.9/projects", json={"name": "my_project"})
    assert response.status_code == 200



import time
from typing import Tuple, Union

import pytest
from nyckel import (
    ClassificationAnnotation,
    ClassificationLabel,
    ClassificationPrediction,
    TextClassificationFunction,
    TextClassificationSample,
)

post_sample_parameter_examples = [
    TextClassificationSample(data="Hi neighbor!", annotation=ClassificationAnnotation(label_name="Nice")),
    ("Hi neighbor!", "Nice"),
    "Hi neighbor!",
]


@pytest.mark.parametrize("post_samples_input", post_sample_parameter_examples)
def test_post_sample_overloading(
    text_classification_function: TextClassificationFunction,
    post_samples_input: Union[TextClassificationSample, Tuple[str, str], str],
) -> None:
    text_classification_function.create_samples([post_samples_input])
    time.sleep(1)
    samples = text_classification_function.list_samples()
    assert len(samples) == 1
    assert samples[0].data == "Hi neighbor!"


def test_samples(text_classification_function: TextClassificationFunction) -> None:
    func = text_classification_function
    label = ClassificationLabel(name="Nice")
    func.create_labels([label])

    sample = TextClassificationSample(data="hello", annotation=ClassificationAnnotation(label_name="Nice"))
    sample_ids = func.create_samples([sample])
    sample_back = func.read_sample(sample_ids[0])
    assert sample_back.data == sample.data

    sample = TextClassificationSample(
        data="hello again", annotation=ClassificationAnnotation(label_name="Nice"), external_id="my_external_id"
    )
    sample_ids = func.create_samples([sample])
    sample_back = func.read_sample(sample_ids[0])
    assert sample_back.data == sample.data
    assert sample_back.external_id == sample.external_id

    # Check that list_samples work
    samples_back = func.list_samples()
    assert len(samples_back) == 2
    assert set([sample.data for sample in samples_back]) == set(["hello", "hello again"])

    # Check delete
    func.delete_samples(sample_ids[0:1])
    time.sleep(1)
    samples_back = func.list_samples()
    assert len(samples_back) == 1
    assert samples_back[0].data == "hello"


def test_end_to_end(text_classification_function: TextClassificationFunction) -> None:
    func = text_classification_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)

    samples = [
        TextClassificationSample(data="hello", annotation=ClassificationAnnotation(label_name="Nice")),
        TextClassificationSample(data="hi", annotation=ClassificationAnnotation(label_name="Nice")),
        TextClassificationSample(data="Good bye", annotation=ClassificationAnnotation(label_name="Boo")),
        TextClassificationSample(data="I'm leaving", annotation=ClassificationAnnotation(label_name="Boo")),
        TextClassificationSample(data="Hi again"),
    ]

    func.create_samples(samples)

    while not func.has_trained_model():
        print("-> No trained model yet. Sleeping 1 sec...")
        time.sleep(1)

    assert isinstance(func.invoke(["Howdy"])[0], ClassificationPrediction)

    assert len(func.invoke(["Hej", "Hola", "Gruezi"])) == 3

    returned_samples = func.list_samples()
    assert len(samples) == len(returned_samples)
    assert isinstance(returned_samples[0], TextClassificationSample)



import time

from nyckel import Credentials
from nyckel.functions.tags import image_tags, text_tags
from nyckel.functions.tags.tags_function_handler import TagsFunctionHandler
from nyckel.functions.utils import strip_nyckel_prefix


class TagsFunctionFactory:

    def __init__(self) -> None:
        self.function_type_by_input = {
            "Text": text_tags.TextTagsFunction,
            "Image": image_tags.ImageTagsFunction,
            # "Tabular": tabular_classification.TabularClassificationFunction,
        }

    @classmethod
    def load(self, function_id: str, credentials: Credentials):
        function_handler = TagsFunctionHandler(function_id, credentials)
        input_modality = function_handler.get_input_modality()
        return self.function_type_by_input[input_modality](function_id, credentials)

    def create(self, name: str, function_input: str, credentials: Credentials):
        def post_function() -> str:
            url = f"{credentials.server_url}/v1/functions"
            response = session.post(url, json={"input": function_input, "output": "Tags", "name": name})
            assert response.status_code == 200, f"Something went wrong when creating function: {response.text}"
            prefixed_function_id = response.json()["id"]
            function_id = strip_nyckel_prefix(prefixed_function_id)
            return function_id

        def hold_until_available(function_id: str) -> None:
            # Before returning, make sure the new function is available via the API.
            timeout_seconds = 5
            t0 = time.time()
            function_is_available = False
            while not function_is_available:
                if time.time() - t0 > timeout_seconds:
                    raise ValueError("Something went wrong when creating function.")
                time.sleep(0.25)
                url = f"{credentials.server_url}/v1/functions/{function_id}"
                response = session.get(url)
                function_is_available = response.status_code == 200

        session = credentials.get_session()
        function_id = post_function()
        hold_until_available(function_id)

        print(f"-> Created function {name} with id: {function_id}")
        return self.function_type_by_input[function_input](function_id, credentials)



import os
import time
import warnings
from typing import Dict, Tuple, Union

import pytest
from nyckel import (
    ClassificationAnnotation,
    ClassificationLabel,
    ClassificationPrediction,
    ImageDecoder,
    TabularClassificationFunction,
    TabularClassificationSample,
)
from nyckel.functions.classification.classification import TabularFunctionField


def test_end_to_end(tabular_classification_function: TabularClassificationFunction) -> None:
    func: TabularClassificationFunction = tabular_classification_function

    func.create_labels([ClassificationLabel(name="happy"), ClassificationLabel(name="sad")])
    func.create_fields(
        [
            TabularFunctionField(name="firstname", type="Text"),
            TabularFunctionField(name="lastname", type="Text"),
            TabularFunctionField(name="mugshot", type="Image"),
        ]
    )
    local_image_file = os.path.abspath("tests/fixtures/flower.jpg")
    func.create_samples(
        [
            TabularClassificationSample(
                data={"firstname": "Adam", "lastname": "Art", "mugshot": local_image_file},
                annotation=ClassificationAnnotation(label_name="happy"),
            ),
            TabularClassificationSample(
                data={"firstname": "Bo", "lastname": "Busy"}, annotation=ClassificationAnnotation(label_name="happy")
            ),
            TabularClassificationSample(
                data={"firstname": "Carl", "lastname": "Carrot"}, annotation=ClassificationAnnotation(label_name="sad")
            ),
            TabularClassificationSample(
                data={"firstname": "Dick", "lastname": "Denali"}, annotation=ClassificationAnnotation(label_name="sad")
            ),
        ]
    )

    fields = func.list_fields()
    assert len(fields) == 3
    assert set([field.name for field in fields]) == set(["firstname", "lastname", "mugshot"])
    time.sleep(2)
    assert len(func.list_samples()) == 4
    assert len(func.list_labels()) == 2

    while not func.has_trained_model():
        print("-> No trained model yet. Sleeping 1 sec...")
        time.sleep(1)

    prediction = func.invoke([{"firstname": "Eric", "lastname": "Ebony", "mugshot": local_image_file}])[0]
    assert isinstance(prediction, ClassificationPrediction)

    predictions = func.invoke(
        [{"firstname": "Eric", "lastname": "Ebony"}, {"firstname": "Frank", "lastname": "Froggy"}]
    )
    assert isinstance(predictions[0], ClassificationPrediction)


def test_field_creation(tabular_classification_function: TabularClassificationFunction) -> None:
    func: TabularClassificationFunction = tabular_classification_function
    func.create_fields([TabularFunctionField(name="firstname", type="Text")])
    func.create_samples([TabularClassificationSample(data={"firstname": "Adam"})])
    fields = func.list_fields()
    assert len(fields) == 1
    assert fields[0].type == "Text"
    func.create_fields([TabularFunctionField(name="age", type="Number")])
    func.create_samples([TabularClassificationSample(data={"age": 32})])
    fields = func.list_fields()
    assert len(fields) == 2
    age_field = [field for field in fields if field.name == "age"][0]
    assert age_field.type == "Number"

    warnings.filterwarnings("ignore", category=RuntimeWarning)
    # Can't assign a Text value to a Number field. This sample will not be created.
    assert len(func.create_samples([TabularClassificationSample(data={"age": "Twelve"})])) == 0
    warnings.filterwarnings("default", category=FutureWarning)


post_sample_parameter_examples = [
    TabularClassificationSample(
        data={"firstname": "Adam", "lastname": "Abrams"}, annotation=ClassificationAnnotation(label_name="Person")
    ),
    ({"firstname": "Adam", "lastname": "Abrams"}, "Person"),
    {"firstname": "Adam", "lastname": "Abrams"},
]


@pytest.mark.parametrize("post_samples_input", post_sample_parameter_examples)
def test_post_sample_overloading(
    tabular_classification_function: TabularClassificationFunction,
    post_samples_input: Union[TabularClassificationSample, Tuple[Dict, str], Dict],
) -> None:
    tabular_classification_function.create_fields(
        [TabularFunctionField(name="firstname", type="Text"), TabularFunctionField(name="lastname", type="Text")]
    )
    tabular_classification_function.create_samples([post_samples_input])
    time.sleep(1)
    samples = tabular_classification_function.list_samples()
    assert len(samples) == 1
    assert samples[0].data == {"firstname": "Adam", "lastname": "Abrams"}


def test_image_field(tabular_classification_function: TabularClassificationFunction) -> None:
    tabular_classification_function.create_fields(
        [TabularFunctionField(name="firstname", type="Text"), TabularFunctionField(name="mug", type="Image")]
    )

    local_filepath = os.path.abspath("tests/fixtures/flower.jpg")

    tabular_classification_function.create_samples(
        [
            TabularClassificationSample(
                data={"firstname": "Adam", "mug": local_filepath},
                annotation=ClassificationAnnotation(label_name="Person"),
            )
        ]
    )
    time.sleep(1)
    samples = tabular_classification_function.list_samples()
    assert isinstance(samples[0].data["mug"], str)
    img = ImageDecoder().to_image(samples[0].data["mug"])

    # The tabular filed uploader will resize the images to 384x384
    assert max(img.size) == 384



import time

from nyckel import (
    ClassificationLabel,
    ClassificationPrediction,
    TagsAnnotation,
    TextTagsFunction,
    TextTagsSample,
)


def compare_tags_samples(sample1: TextTagsSample, sample2: TextTagsSample) -> bool:
    def remove_negative_annotations(sample: TextTagsSample) -> TextTagsSample:
        if sample.annotation:
            sample.annotation = [a for a in sample.annotation if a.present]
        return sample

    sample1.id = None
    sample2.id = None
    sample1 = remove_negative_annotations(sample1)
    sample2 = remove_negative_annotations(sample2)
    return sample1 == sample2


def test_labels(text_tags_function: TextTagsFunction) -> None:
    func = text_tags_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    label_ids = func.create_labels(labels_to_create)

    assert len(func.list_labels()) == 2

    func.delete_labels(label_ids[0:1])
    time.sleep(1)
    assert len(func.list_labels()) == 1


def test_unlabeled_samples(text_tags_function: TextTagsFunction) -> None:
    func = text_tags_function

    samples = ["This is a nice text", "This is a bad text"]

    func.create_samples(samples)
    time.sleep(1)
    assert len(func.list_samples()) == 2


def test_labeled_samples(text_tags_function: TextTagsFunction) -> None:
    func = text_tags_function

    samples = [
        TextTagsSample(data="This is a nice text", annotation=[TagsAnnotation(label_name="Nice", present=True)]),
        TextTagsSample(data="This is a bad text", annotation=[TagsAnnotation(label_name="Bad", present=True)]),
    ]

    sample_ids = func.create_samples(samples)
    time.sleep(1)
    samples_back = func.list_samples()
    assert len(samples_back) == 2
    first_sample = func.read_sample(sample_ids[0])
    assert compare_tags_samples(first_sample, samples[0])


def test_update_annotation(text_tags_function: TextTagsFunction) -> None:
    func = text_tags_function

    samples = [
        TextTagsSample(data="This is a nice text", annotation=[TagsAnnotation(label_name="Nice", present=True)]),
        TextTagsSample(data="This is a bad text", annotation=[TagsAnnotation(label_name="Bad", present=True)]),
    ]

    func.create_samples(samples)
    time.sleep(1)
    samples = func.list_samples()  # type: ignore
    samples[0].annotation = [
        TagsAnnotation(label_name="Nice", present=True),
        TagsAnnotation(label_name="Bad", present=True),
    ]
    func.update_annotation(samples[0])
    time.sleep(1)
    assert samples[0].id is not None
    updated_sample = func.read_sample(samples[0].id)
    assert compare_tags_samples(updated_sample, samples[0])


def test_delete_sample(text_tags_function: TextTagsFunction) -> None:
    func = text_tags_function

    samples = [
        TextTagsSample(data="This is a nice text", annotation=[TagsAnnotation(label_name="Nice", present=True)]),
        TextTagsSample(data="This is a bad text", annotation=[TagsAnnotation(label_name="Bad", present=True)]),
    ]

    sample_ids = func.create_samples(samples)
    time.sleep(1)
    func.delete_samples(sample_ids[0:1])
    time.sleep(1)
    assert len(func.list_samples()) == 1


def test_new_labels(text_tags_function: TextTagsFunction) -> None:

    func = text_tags_function

    func.create_labels([ClassificationLabel(name="label1")])
    time.sleep(0.5)
    sample_1_id = func.create_samples([TextTagsSample(data="sample1", annotation=[TagsAnnotation("label1")])])[0]

    func.create_labels([ClassificationLabel(name="label2")])
    time.sleep(0.5)
    sample_2_id = func.create_samples([TextTagsSample(data="sample2", annotation=[TagsAnnotation("label1")])])[0]

    sample1 = func.read_sample(sample_1_id)

    # When sample1 was created, label2 wasn't present yet.
    # So sample1 doesn't have an "opinion" about that label.
    # This means that it's not present in the annotation field.
    assert len(sample1.annotation) == 1
    assert TagsAnnotation("label1") in sample1.annotation
    assert TagsAnnotation("label2", present=False) not in sample1.annotation
    time.sleep(0.5)
    sample2 = func.read_sample(sample_2_id)
    # When sample2 was created, both labels were present.
    # Since we didn't give an annotation for label2, it was set to not present.
    assert len(sample2.annotation) == 2
    assert TagsAnnotation("label1") in sample2.annotation
    assert TagsAnnotation("label2", present=False) in sample2.annotation


def test_end_to_end(text_tags_function: TextTagsFunction) -> None:
    func = text_tags_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)

    samples = [
        TextTagsSample(data="hello", annotation=[TagsAnnotation("Nice")]),
        TextTagsSample(data="hi", annotation=[TagsAnnotation(label_name="Nice")]),
        TextTagsSample(data="Good bye", annotation=[TagsAnnotation(label_name="Boo")]),
        TextTagsSample(data="I'm leaving", annotation=[TagsAnnotation(label_name="Boo")]),
        TextTagsSample(data="Hi again"),
    ]

    func.create_samples(samples)

    while not func.has_trained_model():
        print("-> No trained model yet. Sleeping 1 sec...")
        time.sleep(1)

    predictions = func.invoke(["Hej", "Hola", "Gruezi"])
    assert len(predictions) == 3
    for prediction in predictions:
        if len(prediction) > 0:
            assert isinstance(prediction[0], ClassificationPrediction)

    returned_samples = func.list_samples()
    assert len(samples) == len(returned_samples)
    assert isinstance(returned_samples[0], TextTagsSample)



import abc
from typing import Dict, List, Sequence, Union

from PIL import Image

from nyckel import (
    ClassificationLabel,
    ClassificationPrediction,
    Credentials,
    ImageSampleData,
    ImageTagsSample,
    NyckelId,
    TagsAnnotation,
    TagsPrediction,
)
from nyckel.functions.classification.image_classification import ImageDecoder, ImageEncoder
from nyckel.functions.classification.label_handler import ClassificationLabelHandler
from nyckel.functions.tags import tags_function_factory
from nyckel.functions.tags.tags import TagsFunctionURLHandler
from nyckel.functions.tags.tags_function_handler import TagsFunctionHandler
from nyckel.functions.tags.tags_sample_handler import TagsSampleHandler
from nyckel.functions.utils import strip_nyckel_prefix
from nyckel.image_processing import ImageResizer


class ImageTagsFunctionInterface(abc.ABC):

    @abc.abstractmethod
    def __init__(self, function_id: str, credentials: Credentials):
        pass

    @property
    @abc.abstractmethod
    def function_id(self) -> NyckelId:
        pass

    @property
    @abc.abstractmethod
    def sample_count(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def label_count(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @classmethod
    @abc.abstractmethod
    def create(cls, name: str, credentials: Credentials) -> "ImageTagsFunction":
        pass

    @abc.abstractmethod
    def delete(self) -> None:
        pass

    @abc.abstractmethod
    def invoke(self, sample_data_list: List[ImageSampleData]) -> List[TagsPrediction]:
        """Invokes the trained function. Raises ValueError if function is not trained"""
        pass

    @abc.abstractmethod
    def has_trained_model(self) -> bool:
        pass

    @abc.abstractmethod
    def create_labels(self, labels: Sequence[Union[ClassificationLabel, str]]) -> List[NyckelId]:
        pass

    @abc.abstractmethod
    def list_labels(self) -> List[ClassificationLabel]:
        pass

    @abc.abstractmethod
    def read_label(self, label_id: NyckelId) -> ClassificationLabel:
        pass

    @abc.abstractmethod
    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        pass

    @abc.abstractmethod
    def delete_labels(self, label_ids: List[NyckelId]) -> None:
        pass

    @abc.abstractmethod
    def create_samples(self, samples: Sequence[Union[ImageTagsSample, ImageSampleData, Image.Image]]) -> List[NyckelId]:
        pass

    @abc.abstractmethod
    def list_samples(self) -> List[ImageTagsSample]:
        pass

    @abc.abstractmethod
    def read_sample(self, sample_id: NyckelId) -> ImageTagsSample:
        pass

    @abc.abstractmethod
    def update_annotation(self, sample: ImageTagsSample) -> None:
        pass

    @abc.abstractmethod
    def delete_samples(self, sample_ids: List[NyckelId]) -> None:
        pass


class ImageTagsFunction(ImageTagsFunctionInterface):
    """
    Example:

    ```py
    from nyckel import Credentials, ImageTagsFunction, ImageTagsSample, TagsAnnotation

    credentials = Credentials(client_id="...", client_secret="...")

    func = ImageTagsFunction.create("ClothingColor", credentials)
    func.create_samples([
        ImageTagsSample(data="t-shirt1.jpg", annotation=[TagsAnnotation("White"), TagsAnnotation("Blue")]),
        ImageTagsSample(data="t=shirt2.jpg", annotation=[TagsAnnotation("Red"), TagsAnnotation("White")]),
        ImageTagsSample(data="jacket.jpg", annotation=[TagsAnnotation("Black")]),
        ImageTagsSample(data="jeans.jpg", annotation=[TagsAnnotation("Blue")]),
    ])
    predictions = func.invoke(["new-jacket.jpg"])
    ```
    """

    def __init__(self, function_id: NyckelId, credentials: Credentials):
        self._function_id = function_id

        self._function_handler = TagsFunctionHandler(function_id, credentials)
        self._label_handler = ClassificationLabelHandler(function_id, credentials)
        self._url_handler = TagsFunctionURLHandler(function_id, credentials.server_url)
        self._sample_handler = TagsSampleHandler(function_id, credentials)
        self._decoder = ImageDecoder()
        self._encoder = ImageEncoder()

        assert self._function_handler.get_input_modality() == "Image"

    @property
    def function_id(self) -> NyckelId:
        return self._function_id

    @property
    def sample_count(self) -> int:
        return self._function_handler.sample_count

    @property
    def label_count(self) -> int:
        return self._function_handler.label_count

    @property
    def name(self) -> str:
        return self._function_handler.get_name()

    @classmethod
    def create(cls, name: str, credentials: Credentials) -> "ImageTagsFunction":
        return tags_function_factory.TagsFunctionFactory().create(name, "Image", credentials)  # type:ignore

    def delete(self) -> None:
        self._function_handler.delete()

    def invoke(self, sample_data_list: List[ImageSampleData]) -> List[TagsPrediction]:
        return self._sample_handler.invoke(sample_data_list, ImageSampleBodyTransformer())  # type: ignore

    def has_trained_model(self) -> bool:
        return self._function_handler.is_trained

    def create_labels(self, labels: Sequence[Union[ClassificationLabel, str]]) -> List[NyckelId]:  # type:ignore
        typed_labels = [
            label if isinstance(label, ClassificationLabel) else ClassificationLabel(name=label)  # type:ignore
            for label in labels
        ]
        return self._label_handler.create_labels(typed_labels)

    def list_labels(self) -> List[ClassificationLabel]:
        return self._label_handler.list_labels(self.label_count)

    def read_label(self, label_id: NyckelId) -> ClassificationLabel:
        return self._label_handler.read_label(label_id)

    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        return self._label_handler.update_label(label)

    def delete_labels(self, label_ids: List[NyckelId]) -> None:
        return self._label_handler.delete_labels(label_ids)

    def create_samples(self, samples: Sequence[Union[ImageTagsSample, ImageSampleData, Image.Image]]) -> List[NyckelId]:
        typed_samples = self._wrangle_post_samples_input(samples)
        typed_samples = self._strip_label_names(typed_samples)
        self._create_labels_as_needed(typed_samples)
        return self._sample_handler.create_samples(typed_samples, ImageSampleBodyTransformer())

    def _wrangle_post_samples_input(
        self, samples: Sequence[Union[ImageTagsSample, ImageSampleData]]
    ) -> List[ImageTagsSample]:
        typed_samples: List[ImageTagsSample] = []
        for sample in samples:
            if isinstance(sample, str):
                typed_samples.append(ImageTagsSample(data=sample))
            elif isinstance(sample, Image.Image):
                typed_samples.append(ImageTagsSample(data=self._encoder.to_base64(sample)))
            elif isinstance(sample, ImageTagsSample):
                typed_samples.append(sample)
            else:
                raise ValueError(f"Unknown sample type: {type(sample)}")
        return typed_samples

    def _strip_label_names(self, samples: List[ImageTagsSample]) -> List[ImageTagsSample]:
        for sample in samples:
            if sample.annotation:
                for entry in sample.annotation:
                    entry.label_name = entry.label_name.strip()
        return samples

    def _create_labels_as_needed(self, samples: List[ImageTagsSample]) -> None:
        existing_labels = self._label_handler.list_labels(None)
        existing_label_names = {label.name for label in existing_labels}
        new_label_names: set = set()
        for sample in samples:
            if sample.annotation:
                new_label_names |= {annotation.label_name for annotation in sample.annotation}
        missing_label_names = new_label_names - existing_label_names
        missing_labels = [ClassificationLabel(name=label_name) for label_name in missing_label_names]
        if len(missing_labels) > 0:
            self._label_handler.create_labels(missing_labels)

    def list_samples(self) -> List[ImageTagsSample]:
        samples_dict_list = self._sample_handler.list_samples(self.sample_count)
        labels = self._label_handler.list_labels(None)
        label_name_by_id = {label.id: label.name for label in labels}

        return [self._sample_from_dict(entry, label_name_by_id) for entry in samples_dict_list]  # type: ignore

    def _sample_from_dict(self, sample_dict: Dict, label_name_by_id: Dict) -> ImageTagsSample:
        if "annotation" in sample_dict:
            annotation = [
                TagsAnnotation(
                    label_name=label_name_by_id[strip_nyckel_prefix(entry["labelId"])],
                    present=entry["present"],
                )
                for entry in sample_dict["annotation"]
            ]
        else:
            annotation = None

        if "prediction" in sample_dict:
            prediction = [
                ClassificationPrediction(
                    confidence=entry["confidence"],
                    label_name=label_name_by_id[strip_nyckel_prefix(entry["labelId"])],
                )
                for entry in sample_dict["prediction"]
            ]
        else:
            prediction = None

        return ImageTagsSample(
            id=strip_nyckel_prefix(sample_dict["id"]),
            data=sample_dict["data"],
            external_id=sample_dict["externalId"] if "externalId" in sample_dict else None,
            annotation=annotation,
            prediction=prediction,
        )

    def read_sample(self, sample_id: NyckelId) -> ImageTagsSample:
        sample_dict = self._sample_handler.read_sample(sample_id)

        labels = self._label_handler.list_labels(None)
        label_name_by_id = {strip_nyckel_prefix(label.id): label.name for label in labels}  # type: ignore

        return self._sample_from_dict(sample_dict, label_name_by_id)  # type: ignore

    def update_annotation(self, sample: ImageTagsSample) -> None:
        self._sample_handler.update_annotation(sample)

    def delete_samples(self, sample_ids: List[NyckelId]) -> None:
        self._sample_handler.delete_samples(sample_ids)


class ImageSampleBodyTransformer:

    def __init__(self):
        self._decoder = ImageDecoder()
        self._encoder = ImageEncoder()
        self._resizer = ImageResizer()

    def __call__(self, sample_data: ImageSampleData) -> str:
        """Resizes if needed and encodes the sample data as a URL or dataURI."""
        if self._is_nyckel_owned_url(sample_data):
            # If the input points to a Nyckel S3 bucket, we know that the image is processed and verified.
            # In that case, we just point back to that URL.
            return sample_data

        if self._decoder.looks_like_url(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        if self._decoder.looks_like_local_filepath(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        if self._decoder.looks_like_data_uri(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        raise ValueError(f"Can't parse input sample.data={sample_data}")

    def _is_nyckel_owned_url(self, sample_data: str) -> bool:
        return sample_data.startswith("https://s3.us-west-2.amazonaws.com/nyckel.server.")



from .auth import Credentials  # noqa: F401, I001
from .auth import Credentials as OAuth2Renewer  # For backwards compatibility. # noqa: F401
from .auth import Credentials as User  # For backwards compatibility. # noqa: F401
from .data_classes import NyckelId  # noqa: F401
from .functions.classification.classification import (
    ClassificationAnnotation,  # noqa: F401
    ClassificationFunction,  # noqa: F401
    ClassificationLabel,  # noqa: F401
    ClassificationPrediction,  # noqa: F401
    ClassificationSample,  # noqa: F401
    ImageClassificationSample,  # noqa: F401
    ImageSampleData,  # noqa: F401
    LabelName,  # noqa: F401
    TabularClassificationSample,  # noqa: F401
    TabularFieldKey,  # noqa: F401
    TabularFieldValue,  # noqa: F401
    TabularFunctionField,  # noqa: F401
    TabularSampleData,  # noqa: F401
    TextClassificationSample,  # noqa: F401
    TextSampleData,  # noqa: F401
)

from .functions.tags.tags import (
    ImageTagsSample,  # noqa: F401
    TabularTagsSample,  # noqa: F401
    TagsAnnotation,  # noqa: F401
    TagsPrediction,  # noqa: F401
    TextTagsSample,  # noqa: F401
)

from .functions.classification.factory import ClassificationFunctionFactory  # noqa: F401
from .functions.classification.image_classification import (
    ImageClassificationFunction,  # noqa: F401
    ImageDecoder,  # noqa: F401
    ImageEncoder,  # noqa: F401
)
from .functions.classification.tabular_classification import TabularClassificationFunction  # noqa: F401
from .functions.classification.text_classification import TextClassificationFunction  # noqa: F401
from .functions.tags.text_tags import TextTagsFunction  # noqa: F401
from .functions.tags.image_tags import ImageTagsFunction  # noqa: F401



import os
import time
from typing import Iterator

from nyckel import ImageClassificationFunction
from nyckel.functions.classification.image_classification import ImageDecoder

FULLY_TRANSPARENT_PIXEL_THAT_SHOULD_BECOME_WHITE = (10, 10)


def test_white_background(image_classification_function: Iterator[ImageClassificationFunction]):
    func: ImageClassificationFunction = image_classification_function
    local_filepath = os.path.abspath("tests/fixtures/mixed-alpha-background.png")
    sample_id = func.create_samples([local_filepath])[0]
    time.sleep(1)
    sample_back = func.read_sample(sample_id)
    img = ImageDecoder().to_image(sample_back.data)
    assert img.getpixel(FULLY_TRANSPARENT_PIXEL_THAT_SHOULD_BECOME_WHITE) == (255, 255, 255)



import abc
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from nyckel import Credentials, NyckelId

TextSampleData = str


ImageSampleData = str
"""DataUri, Url, or local filepath."""


TabularFieldValue = Union[str, float]
TabularFieldKey = str
TabularSampleData = Dict[TabularFieldKey, TabularFieldValue]

LabelName = str


@dataclass
class ClassificationLabel:
    name: LabelName
    id: Optional[NyckelId] = None
    description: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None


@dataclass
class ClassificationPrediction:
    label_name: LabelName
    confidence: float


@dataclass
class ClassificationAnnotation:
    label_name: str


@dataclass
class TabularFunctionField:
    name: str
    type: str
    """Type must be Number, Text or Image"""
    id: Optional[NyckelId] = None

    def __post_init__(self) -> None:
        assert self.type in ["Number", "Text", "Image"]


@dataclass
class ImageClassificationSample:
    data: ImageSampleData
    id: Optional[NyckelId] = None
    external_id: Optional[str] = None
    annotation: Optional[ClassificationAnnotation] = None
    prediction: Optional[ClassificationPrediction] = None


@dataclass
class TextClassificationSample:
    data: TextSampleData
    id: Optional[NyckelId] = None
    external_id: Optional[str] = None
    annotation: Optional[ClassificationAnnotation] = None
    prediction: Optional[ClassificationPrediction] = None


@dataclass
class TabularClassificationSample:
    data: TabularSampleData
    id: Optional[NyckelId] = None
    external_id: Optional[str] = None
    annotation: Optional[ClassificationAnnotation] = None
    prediction: Optional[ClassificationPrediction] = None


ClassificationSample = Union[TextClassificationSample, ImageClassificationSample, TabularClassificationSample]


class ClassificationFunction(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def create(cls, name: str, credentials: Credentials) -> "ClassificationFunction":
        pass

    @abc.abstractmethod
    def delete(self) -> None:
        pass

    @abc.abstractmethod
    def __init__(self, function_id: str, credentials: Credentials):
        pass

    @property
    @abc.abstractmethod
    def function_id(self) -> NyckelId:
        pass

    @property
    @abc.abstractmethod
    def sample_count(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def label_count(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @abc.abstractmethod
    def invoke(self, sample_data_list: List[str], model_id: str) -> List[ClassificationPrediction]:
        """Invokes the trained function. Raises ValueError if function is not trained"""
        pass

    @abc.abstractmethod
    def has_trained_model(self) -> bool:
        pass

    @abc.abstractmethod
    def create_labels(self, labels: List[ClassificationLabel]) -> List[NyckelId]:
        pass

    @abc.abstractmethod
    def list_labels(self) -> List[ClassificationLabel]:
        pass

    @abc.abstractmethod
    def read_label(self, label_id: NyckelId) -> ClassificationLabel:
        pass

    @abc.abstractmethod
    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        pass

    @abc.abstractmethod
    def delete_labels(self, label_ids: List[NyckelId]) -> None:
        pass

    @abc.abstractmethod
    def create_samples(self, samples: List[ClassificationSample]) -> List[NyckelId]:
        pass

    @abc.abstractmethod
    def list_samples(self) -> List[ClassificationSample]:
        pass

    @abc.abstractmethod
    def read_sample(self, sample_id: NyckelId) -> ClassificationSample:
        pass

    @abc.abstractmethod
    def update_annotation(self, sample: ClassificationSample) -> None:
        pass

    @abc.abstractmethod
    def delete_samples(self, sample_ids: List[NyckelId]) -> None:
        pass


class ClassificationFunctionURLHandler:
    def __init__(self, function_id: NyckelId, server_url: str):
        self._function_id = function_id
        self._server_url = server_url

    @property
    def train_page(self) -> str:
        return f"{self._server_url}/console/functions/{self._function_id}/train"

    def api_endpoint(self, path: Optional[str] = None, api_version: str = "v1", query_str: Optional[str] = None) -> str:
        endpoint = f"{self._server_url}/{api_version}/functions/{self._function_id}"
        if path:
            endpoint += f"/{path}"
        if query_str:
            endpoint += f"?{query_str}"
        return endpoint



import concurrent.futures
import warnings
from json import JSONDecodeError
from typing import Callable, Dict, List, Optional, Tuple

import requests
from requests.adapters import HTTPAdapter, Retry
from tqdm import tqdm

from nyckel.config import NBR_CONCURRENT_REQUESTS


class ParallelPoster:
    def __init__(
        self,
        session: requests.Session,
        endpoint: str,
        progress_bar: Optional[tqdm] = None,
        body_transformer: Callable = lambda x: x,
    ):
        """last input 'body_transformer' intended for image functions,
        where we want to resize and such at the very last instance before posting.
        This is to avoid blowing up the memory for large sets of images to post."""
        self._session = session
        self._endpoint = endpoint
        self._body_transformer = body_transformer
        if progress_bar is not None:
            self.progress_bar = progress_bar

        else:
            self.progress_bar = tqdm("Posting", ncols=80)

    def _post_as_json(self, data: Dict) -> requests.Response:
        response = self._session.post(self._endpoint, json=self._body_transformer(data))
        return response

    def refresh_session(self, session: requests.Session) -> None:
        self._session = session

    def __call__(self, bodies: List[Dict]) -> List[requests.Response]:
        if len(bodies) == 0:
            return []
        responses = [requests.Response()] * len(bodies)
        n_workers = min(len(bodies), NBR_CONCURRENT_REQUESTS)

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
            index_by_future = {executor.submit(self._post_as_json, body): index for index, body in enumerate(bodies)}
            for future in concurrent.futures.as_completed(index_by_future):
                index = index_by_future[future]
                body = bodies[index]
                self.progress_bar.update(1)
                try:
                    response = future.result()
                    if response.status_code not in [200, 409]:
                        warnings.warn(
                            f"Posting {body} to {self._endpoint} failed with {response.status_code=} {response.text=}",
                            RuntimeWarning,
                        )
                    responses[index] = response
                except Exception as e:
                    warnings.warn(f"Posting {body} to {self._endpoint} failed with {e}", RuntimeWarning)

        return responses


class ParallelDeleter:
    def __init__(
        self,
        session: requests.Session,
        endpoint: str,
        desc: Optional[str] = None,
    ):
        self._session = session
        self._endpoint = endpoint
        self._desc = desc

    def _delete_one(self, asset_id: str) -> requests.Response:
        response = self._session.delete(f"{self._endpoint}/{asset_id}")
        return response

    def __call__(self, asset_ids: List[str]) -> List[requests.Response]:
        responses = [requests.Response()] * len(asset_ids)
        n_workers = min(len(asset_ids), NBR_CONCURRENT_REQUESTS)

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
            index_by_future = {
                executor.submit(self._delete_one, asset_id): index for index, asset_id in enumerate(asset_ids)
            }
            for future in tqdm(
                concurrent.futures.as_completed(index_by_future), total=len(asset_ids), desc=self._desc, ncols=80
            ):
                index = index_by_future[future]
                asset_id = asset_ids[index]
                response = future.result()
                if not response.status_code == 200:
                    raise ValueError(
                        f"Error when deleting asset: {self._endpoint}/{asset_id}. {response.status_code=} "
                        f"{response.text=}"
                    )
                responses[index] = response
        return responses


class SequentialGetter:
    def __init__(self, session: requests.Session, endpoint: str):
        self._session = session
        self._endpoint = endpoint

    def __call__(self, progress_bar: Optional[tqdm] = None) -> List[Dict]:
        base_url, slug = self._get_base_url()

        resp = self._session.get(base_url + slug)
        if not resp.status_code == 200:
            raise RuntimeError(f"GET from {base_url+slug} failed with {resp.status_code}, {resp.text}.")

        resource_list = resp.json()
        if progress_bar is not None:
            progress_bar.update(len(resource_list))

        while "next" in resp.links:
            slug = resp.links["next"]["url"]
            resp = self._session.get(base_url + slug)
            if not resp.status_code == 200:
                raise RuntimeError(f"GET from {base_url+slug} failed with {resp.status_code}, {resp.text}.")
            try:
                this_resource_list = resp.json()
            except JSONDecodeError as e:
                print(f"Failed to decode json from {base_url+slug}")
                raise e
            resource_list.extend(this_resource_list)
            if progress_bar is not None:
                progress_bar.update(len(this_resource_list))

        return resource_list

    def _get_base_url(self) -> Tuple[str, str]:
        if ":5000" in self._endpoint:
            # Running locally
            parts = self._endpoint.split(":5000")
            base_url, slug = parts
            base_url += ":5000"
        else:
            parts = self._endpoint.split(".com")
            base_url, slug = parts
            base_url += ".com"
        return base_url, slug


def get_session_that_retries() -> requests.Session:
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    session.mount("http://", HTTPAdapter(max_retries=retries))
    return session



from dataclasses import dataclass
from typing import List, Optional, Sequence

from nyckel import NyckelId
from nyckel.functions.classification.classification import (
    ClassificationPrediction,
    ImageSampleData,
    TabularSampleData,
    TextSampleData,
)


@dataclass
class TagsAnnotation:
    label_name: str
    present: bool = True


TagsPrediction = Sequence[ClassificationPrediction]


@dataclass
class ImageTagsSample:
    data: ImageSampleData
    id: Optional[NyckelId] = None
    external_id: Optional[str] = None
    annotation: Optional[List[TagsAnnotation]] = None
    prediction: Optional[List[ClassificationPrediction]] = None


@dataclass
class TextTagsSample:
    data: TextSampleData
    id: Optional[NyckelId] = None
    external_id: Optional[str] = None
    annotation: Optional[List[TagsAnnotation]] = None
    prediction: Optional[List[ClassificationPrediction]] = None


@dataclass
class TabularTagsSample:
    data: TabularSampleData
    id: Optional[NyckelId] = None
    external_id: Optional[str] = None
    annotation: Optional[List[TagsAnnotation]] = None
    prediction: Optional[List[ClassificationPrediction]] = None


class TagsFunctionURLHandler:
    def __init__(self, function_id: NyckelId, server_url: str):
        self._function_id = function_id
        self._server_url = server_url

    @property
    def train_page(self) -> str:
        return f"{self._server_url}/console/functions/{self._function_id}/train"

    def api_endpoint(self, path: Optional[str] = None, api_version: str = "v1", query_str: Optional[str] = None) -> str:
        endpoint = f"{self._server_url}/{api_version}/functions/{self._function_id}"
        if path:
            endpoint += f"/{path}"
        if query_str:
            endpoint += f"?{query_str}"
        return endpoint
